#! /usr/bin/python2

# hal_touchoff_helper defines the following hal pins:
#
# input pins:
#     hal_touchoff_helper.[AXIS].zero
#         on a rising edge will use 'G10 L20' to zero the axis
#     hal_touchoff_helper.[AXIS].half
#         on a rising edge will use 'G10 L20' to set the axis to half the
#         current (relative) value
#     hal_touchoff_helper.[AXIS].set
#         on a rising edge will use 'G10 L20' to set the axis to a user entered
#         value (will prompt to enter the value)
#     hal_touchoff_helper.[CSYS].select
#         on a rising edge will issue to corresponding gcode to select the
#         coordinate system
#     hal_touchoff_helper.prompt
#         when TRUE; prompt the user for confirmation when zeroing or halving an
#         axis
#
# output pins:
#     hal_touchoff_helper.[CSYS].active
#         TRUE when the coordinate system is currently selected
#         (FALSE otherwise)
#     hal_touchoff_helper.G5x.index
#         is the index of the current coordinate system
#         (1 for G54, 2 for G55, ..., 9 for G59.3)
#     hal_touchoff_helper.G5x.name
#         is the 'name' of the current coordinate system
#         (54 for G54, 55 for G55, ..., 59.3 for G59.3)
#
# [AXIS] is one of x, y, z, a, b, c, u, v, w
# [CSYS] is one of G54, G55, G56, G57, G58, G59, G59_1, G59_2, G59_3
#
# a rising edge is defined as a transition of the input pin from FALSE to TRUE.
# you'll probably want to the debounce hal module if you wish to attach these
# pins to physical buttons.

import linuxcnc, hal, gcode
import Tkinter, nf, rs274.options
import gettext, os, sys, shutil, tempfile, atexit

BASE = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), ".."))
gettext.install("linuxcnc", localedir=os.path.join(BASE, "share", "locale"), unicode=True)

################################################################################

# GUI stuff taken from src/emc/usr_intf/axis/scripts/axis.py
# prompt_areyousure() and prompt_float() have two tweaks:
# - don't call wm_transient(root_window)
#   as the root_window isn't shown this would cause the dialog to not be shown
# - call self.t.wait_visibility()
#   otherwise self.t.grab_set() will fail when the corresponding hal pin is wired up to a vcp
# prompt_touchoff() is largely different (but still derivative)

root_window = Tkinter.Tk(className="hal_touchoff_helper")
dpi_value = root_window.winfo_fpixels('1i')
root_window.tk.call('tk', 'scaling', '-displayof', '.', dpi_value / 72.0)
root_window.wm_protocol("WM_DELETE_WINDOW", root_window.wm_withdraw)
root_window.withdraw()
nf.start(root_window)
nf.makecommand(root_window, "_", _)
rs274.options.install(root_window)

path_inifile = os.environ.get("INI_FILE_NAME", "/dev/null")
inifile = linuxcnc.ini(path_inifile)
lathe = bool(inifile.find("DISPLAY", "LATHE"))

def remove_tempdir(t):
    shutil.rmtree(t)
tempdir = tempfile.mkdtemp()
atexit.register(remove_tempdir, tempdir)

class DummyCanon:
    def comment(*args): pass
    def next_line(*args): pass
    def set_g5x_offset(*args): pass
    def set_g92_offset(*args): pass
    def set_xy_rotation(*args): pass
    def get_external_angular_units(self): return 1.0
    def get_external_length_units(self): return 1.0
    def set_plane(*args): pass
    def get_axis_mask(self): return 7
    def get_tool(self, tool):
        return tool, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0
    def set_feed_rate(self, rate): pass

    def user_defined_function(self, m, p, q):
        self.number = p

def parse_gcode_expression(e):
    f = os.path.devnull
    canon = DummyCanon()

    parameter = inifile.find("RS274NGC", "PARAMETER_FILE")
    temp_parameter = os.path.join(tempdir, os.path.basename(parameter))
    shutil.copy(parameter, temp_parameter)
    canon.parameter_file = temp_parameter

    result, seq = gcode.parse("", canon, "M199 P["+e+"]", "M2")
    if result > gcode.MIN_ERROR: return False, gcode.strerror(result)
    return True, canon.number

class _prompt_areyousure:
    """ Prompt for a question, user can enter yes or no """
    def __init__(self, title, text):
        t = self.t = Tkinter.Toplevel(root_window, padx=7, pady=7)
        t.wm_title(title)
        #t.wm_transient(root_window)
        t.wm_resizable(0, 0)
        self.status=False
        m = Tkinter.Message(t, text=text, aspect=500, anchor="w", justify="left")
        self.w = w = Tkinter.StringVar(t)
        l = Tkinter.Message(t, textvariable=w, justify="left", anchor="w",
                aspect=500)
        self.buttons = f = Tkinter.Frame(t)
        self.ok = Tkinter.Button(f, text=_("Ok"), command=self.do_ok, width=10,height=1,padx=0,pady=.25, default="active")
        self.cancel = Tkinter.Button(f, text=_("Cancel"), command=self.do_cancel, width=10,height=1,padx=0,pady=.25, default="normal")
        t.wm_protocol("WM_DELETE_WINDOW", self.cancel.invoke)
        t.bind("<Return>", lambda event: (self.ok.flash(), self.ok.invoke()))
        t.bind("<KP_Enter>", lambda event: (self.ok.flash(), self.ok.invoke()))
        t.bind("<space>", lambda event: (self.ok.flash(), self.ok.invoke()))
        t.bind("<Escape>", lambda event: (self.cancel.flash(), self.cancel.invoke()))

        m.pack(side="top", anchor="w")
        l.pack(side="top", anchor="w", fill="x", expand=1)
        f.pack(side="bottom", anchor="e")
        self.ok.pack(side="left", padx=3, pady=3)
        self.cancel.pack(side="left", padx=3, pady=3)

    def do_ok(self):
        self.status=True
        self.t.destroy()

    def do_cancel(self):
        self.status=False
        self.t.destroy()

    def result(self):
        return self.status

    def run(self):
        self.t.wait_visibility()
        self.t.grab_set()
        self.t.wait_window()
        try:
            self.t.destroy()
        except Tkinter.TclError:
            pass
        return self.result()

def prompt_areyousure(title, text):
    t = _prompt_areyousure(title, text)
    return t.run()

class _prompt_float:
    """ Prompt for a G-code floating point expression """
    def __init__(self, title, text, default, unit_str=''):
        self.unit_str = unit_str
        t = self.t = Tkinter.Toplevel(root_window, padx=7, pady=7)
        t.wm_title(title)
        #t.wm_transient(root_window)
        t.wm_resizable(0, 0)
        self.m = m = Tkinter.Message(t, text=text, aspect=500, anchor="w", justify="left")
        self.v = v = Tkinter.StringVar(t)
        self.vv = vv = Tkinter.DoubleVar(t)
        self.u = u = Tkinter.BooleanVar(t)
        self.w = w = Tkinter.StringVar(t)
        l = Tkinter.Message(t, textvariable=w, justify="left", anchor="w",
                aspect=500)
        v.set(default)
        self.e = e = Tkinter.Entry(t, textvariable=v)
        self.buttons = f = Tkinter.Frame(t)
        self.ok = Tkinter.Button(f, text=_("OK"), command=self.do_ok, width=10,height=1,padx=0,pady=.25, default="active")
        self.cancel = Tkinter.Button(f, text=_("Cancel"), command=self.do_cancel, width=10,height=1,padx=0,pady=.25, default="normal")
        v.trace("w", self.check_valid)
        t.wm_protocol("WM_DELETE_WINDOW", self.cancel.invoke)
        t.bind("<Return>", lambda event: (self.ok.flash(), self.ok.invoke()))
        t.bind("<KP_Enter>", lambda event: (self.ok.flash(), self.ok.invoke()))
        t.bind("<Escape>", lambda event: (self.cancel.flash(), self.cancel.invoke()))

        m.pack(side="top", anchor="w")
        e.pack(side="top", anchor="e")
        l.pack(side="top", anchor="w", fill="x", expand=1)
        f.pack(side="bottom", anchor="e")
        self.ok.pack(side="left", padx=3, pady=3)
        self.cancel.pack(side="left", padx=3, pady=3)

    def set_text(self, text):
        self.m.configure(text=text)

    def do_ok(self):
        self.u.set(True)
        self.t.destroy()

    def do_cancel(self):
        self.u.set(False)
        self.t.destroy()

    def check_valid(self, *args):
        v = self.v.get()

        st = 0
        ok = 1

        if "#" in v:
            ok = 0
            self.w.set("Variables may not be used here")

        if ok:
            for ch in v:
                if ch == "[": st += 1
                elif ch == "]": st -= 1
                if st < 0:
                    ok = 0
                    self.w.set("Right bracket without matching left bracket")
                    break
            if st != 0:
                self.w.set("Left bracket without matching right bracket")
                ok = 0

        if ok:
            ok, value = parse_gcode_expression(v)
            if ok:
                self.w.set("= %f%s" % (value, self.unit_str))
                self.vv.set(value)
            else:
                self.w.set(value)

        if ok:
            self.ok.configure(state="normal")
        else:
            self.ok.configure(state="disabled")

    def do_focus(self):
        if not self.e.winfo_viewable():
            self._after = self.t.after(10, self.do_focus)
        else:
            self.e.focus()
            self.e.selection_range(0, "end")
            self._after = None

    def result(self):
        if self.u.get(): return self.vv.get()
        return None

    def run(self):
        self.t.wait_visibility()
        self.t.grab_set()
        self._after = self.t.after_idle(self.do_focus)
        self.t.wait_window()
        if self._after is not None:
            self.t.after_cancel(self._after)
        try:
            self.t.destroy()
        except Tkinter.TclError:
            pass
        return self.result()

def prompt_float(title, text, default, unit_str):
    t = _prompt_float(title, text, default, unit_str)
    return t.run()

def prompt_touchoff(title, text, axis):
    linear_axis = axis in "xyzuvw"
    if linear_axis:
        if 200 in s.gcodes: unit_str = " " + _("in")
        if 210 in s.gcodes: unit_str = " " + _("mm")
        if lathe and axis == "x":
            if 80 in s.gcodes:
                unit_str += _(" radius")
            else:
                unit_str += _(" diameter")
    else:
        unit_str = _(u"\xb0")
    return prompt_float(title, text, 0, unit_str)

################################################################################

def mdi_ready():
    return not s.estop and s.enabled and (s.interp_state == linuxcnc.INTERP_IDLE)

def mdi(gcode):
    old_mode = s.task_mode
    c.mode(linuxcnc.MODE_MDI)
    c.wait_complete()
    c.mdi(gcode)
    c.wait_complete()
    c.mode(old_mode)
    c.wait_complete()

def zero(axis):
    if rising_edge('%s.zero' % axis) and mdi_ready():
        if h["prompt"] and prompt_areyousure("Touch Off", "Zero %s axis?" % axis.upper()) == False:
            return
        mdi('G10 L20 P0 %s0' % axis)

def half(axis):
    if rising_edge('%s.half' % axis) and mdi_ready():
        if h["prompt"] and prompt_areyousure("Touch Off", "Half %s axis value?" % axis.upper()) == False:
            return
        if axis == 'x': mdi('G10 L20 P0 X[#5420/2]')
        if axis == 'y': mdi('G10 L20 P0 Y[#5421/2]')
        if axis == 'z': mdi('G10 L20 P0 Z[#5422/2]')
        if axis == 'a': mdi('G10 L20 P0 A[#5423/2]')
        if axis == 'b': mdi('G10 L20 P0 B[#5424/2]')
        if axis == 'c': mdi('G10 L20 P0 C[#5425/2]')
        if axis == 'u': mdi('G10 L20 P0 U[#5426/2]')
        if axis == 'v': mdi('G10 L20 P0 V[#5427/2]')
        if axis == 'w': mdi('G10 L20 P0 W[#5428/2]')

def set_(axis):
    if rising_edge('%s.set' % axis) and mdi_ready():
        value = prompt_touchoff("Touch Off", "Set %s axis value" % axis.upper(), axis)
        if value:
            mdi("G10 L20 P0 %s%12f" % (axis, value))

def select_work_offset(work_offset):
    # work_offset is something like 'G54' or 'G59_2'
    # - work_offset is used as is for the hal pin name,
    # - but needs tweaking to become gcode
    if rising_edge('%s.select' % work_offset) and mdi_ready():
        mdi(work_offset.replace('_', '.'))

def rising_edge(pin):
    old = old_pin[pin]
    new = h[pin]
    old_pin[pin] = new
    return new and not old

def addpin_for_edge_detection(pin):
    h.newpin(pin, hal.HAL_BIT, hal.HAL_IN)
    old_pin[pin] = False

################################################################################

axes = 'xyzabcuvw'
# work_offsets values are (linuxcnc.stat.g5x_index, CSYS, G5x.name)
work_offsets = [ 
    (1, 'G54', 54),
    (2, 'G55', 55),
    (3, 'G56', 56),
    (4, 'G57', 57),
    (5, 'G58', 58),
    (6, 'G59', 59),
    (7, 'G59_1', 59.1),
    (8, 'G59_2', 59.2),
    (9, 'G59_3', 59.3)
]
old_pin = {}

s = linuxcnc.stat()
c = linuxcnc.command()


# create hal component
h = hal.component('hal_touchoff_helper')
# input pins
for axis in axes:
    addpin_for_edge_detection('%s.zero' % axis)
    addpin_for_edge_detection('%s.half' % axis)
    addpin_for_edge_detection('%s.set' % axis)
for index, csys, name in work_offsets:
    addpin_for_edge_detection('%s.select' % csys)
h.newpin("prompt", hal.HAL_BIT, hal.HAL_IN)
# output pins
for index, csys, name in work_offsets:
    h.newpin('%s.active' % csys, hal.HAL_BIT, hal.HAL_OUT)
h.newpin('G5x.index', hal.HAL_S32, hal.HAL_OUT)
h.newpin('G5x.name', hal.HAL_FLOAT, hal.HAL_OUT)
h.ready()


# main loop
try:
    while 1:
        s.poll()

        # check inputs
        for axis in axes:
            zero(axis)
            half(axis)
            set_(axis)
        for index, csys, name in work_offsets:
            select_work_offset(csys)

        # update outputs
        h['G5x.index'] = s.g5x_index
        h['G5x.name'] = 0
        for index, csys, name in work_offsets:
            if s.g5x_index == index:
                h['%s.active' % csys] = True
                h['G5x.name'] = name
            else:
                h['%s.active' % csys] = False

        # allow prompt window to do its thing
        root_window.after(100)
        root_window.update()

except KeyboardInterrupt:
    pass
